#ifndef _QUERY_H
#define _QUERY_H
#include <stdarg.h>
#include "dynarray.h"
#include "timer.h"
#include "dump.hpp"
#include "error.h"


//		Common types
#define SQL_SHORT                          500
#define SQL_LONG                           496
#define SQL_FLOAT                          482
#define SQL_DOUBLE                         480
#define SQL_FIXCHAR                        452
#define SQL_VARCHAR                        448

//		Extra types
#define SQL_DATE                           510
#define SQL_QUAD                           550

#define SQL_BLOB										520
#define SQL_LTEXT										522


enum UniSqlType {
	UST_NOTYPE = 0,
	UST_CHAR = SQL_FIXCHAR,
	UST_SHORT = SQL_SHORT,
	UST_LONG = SQL_LONG,
	UST_FLOAT = SQL_FLOAT,
	UST_DOUBLE = SQL_DOUBLE,
	UST_DATETIME = SQL_DATE,
	UST_BINARY = SQL_BLOB,
	};

struct UniDT	{
	long days;
	long secs;
	};

struct UniBlob	{
	long size;
	char text[1];
	};

struct UniVar	{
	int uniType;
	void* value;
	UniVar (int _uniType = UST_NOTYPE, void* _value = 0)	{
		uniType = _uniType;
		value = _value;
		}
	};

enum	{
	QUE_SERVER_ERROR,
	QUE_FIELD_NOT_FOUND,
	QUE_CANT_UNPREPARE,
	QUE_BAD_TRANS,
	QUE_BAD_COLUMN,
	QUE_BAD_BLOBID,
	QUE_DISCONNECT_TRANS,
	QUE_BAD_ACTIVATE,
	QUE_CLOSED_CONN,
	QUE_START_TWICE,
	QUE_COMMIT_NO_TRANS,
	QUE_ROLLBACK_NO_TRANS,
	QUE_NO_ROWS,
	QUE_REDEFINE,
	QUE_LONG_NAME,
	QUE_TRUNCATION,
	QUE_TYPE_NOT_SUPPORTED,
	};
extern ErrorClass queryError;
	
#ifndef DB_NO_OLD_SQLDA
struct SQLVAR	{
	short sqltype;
	unsigned short sqllen;
	char* sqldata;
	short* sqlind;
	short sqlname_length;
	char sqlname [30];
	};

struct SQLDA	{
	char sqldaid[8];
	long sqldabc;
	short sqln;
	short sqld;
	SQLVAR sqlvar[1];
	};

#define SQLDA_LENGTH(n)         (sizeof (SQLDA) + (n-1) * sizeof (SQLVAR))

#else
	#include "gds.hpp"
//	class SQLDA;
//	class SQLVAR;
#endif

//			-----------			SQLDA utilities			--------------------
//			______________________________________________________________
#define AVG_FIELD_SIZE  (sizeof(DbRecordField)+8)
#define FIELD_GRAN		16

//		Record buffer element
struct DbRecordField	{
public :
	long length;			//		total length of element
	short nullind;
	short reserved;
	char text[1];
	DbRecordField& operator=(SQLVAR* var);
			//		Fills element with SQLVAR <var> (assuming that there's enough
			//		room after <text> to store <var> value
	};

//		Record buffer consisting of sequence of record buffer elements
class DbRecord : public Dump	{
public :
	int nField;

	DbRecord()	{	nField = 0;	used = 0;	}
	void create(SQLDA* sqlda);		//		Creates record buffer for given <sqlda>
	DbRecordField* get(int i);
	DbRecord& operator<<(SQLVAR* var);		//		adds new element for <var>
	void clear();

	int used;		//		Amount of bytes used by elements
	DbRecordField* extend(int delta);
		//		Ensures that record buffer has enough room to store one more
		//		record element of <delta> bytes, allocates additional space if
		//		necessary, and returns pointer to newly allocated element with
		//		size of <delta> bytes
	};

//		SQLDA structure + Record buffer
class DbSQLDA	{
public :
	Bool hasRecord;			//		If record buffer member is used
	Dump sqlda;
	DbRecord record;
	DbSQLDA()	{	hasRecord = TRUE;	}

	SQLDA* getSQLDA()	{	return (SQLDA*)sqlda.data;	}
	operator SQLDA*()	{	return (SQLDA*)sqlda.data;	}
	
	SQLVAR& operator[](int i)	{	return getSQLDA()->sqlvar[i];	}
	void create(SQLDA* s, Bool withData = FALSE);
				//		Creates copy of SQLDA <s> with record buffer if desired
	void synchronize();
				//		Synchonizes SQLDA structure with record buffer
				//		Must be called before using any library calls with SQLDA
	DbSQLDA& operator<<(SQLVAR* var);
				//		Adds SQLVAR <var> (and data in record buffer too)
	void clear();

	void extend(int delta);		//		ensures that sqlda has enough room to
				//		store <delta> more SQLVARs, allocates additional space
				//		if necessary, adjusts sqlda->sqln member
	};
//			______________________________________________________________



//			--------- 		__Query - basic DB query class	--------------
//			______________________________________________________________

//		Define state flags
#define	DQS_PREPARED		0x0001
#define	DQS_HAS_CURSOR		0x0002
#define	DQS_OPENED			0x0004
#define	DQS_CACHED			0x0008
#define	DQS_HAS_STMT		0x0010
#define	DQS_PROCESSED		0x0020

#define QNAME_LEN 	16
#define MAX_FIELDS_IN_QUERY	512
class __Query {
	friend class StackOfQuery;
	friend class StoredQuery;
public :
	__Query(const char *name = 0);
	virtual void setName(const char *name);

	virtual Result release();
	virtual ~__Query() {};

	const char* getQueryName() const	{	return queryName; };
	virtual char* getCursorName() {	return queryName; };

	virtual char* getStackName() const	{	return "$-default";	}
	static __Query* findByName(const char* name);


	//		Type specifiers :
	//			%B 	Dump* - for binary data
	//			%S 	FString*
	//			%s 	char*
	//			%hd	short by value
	//			%ld	long by value
	//			%hD	short by address
	//			%lD	long by address
	//			%f,%F		float by value, by address
	//			%lf,%lF		double by value, by address
	//			%D		Date*
	//			%Q		DbBlobId* (GDS_QUAD*)
	//			%e, %he, %le - enums stored in database like 'AB' (passed by value)
	//			%E, %hEe, %lE - enums passed by address
	//			%i, &I - short in database, int in process
	//			%U - text dump
	//			%R - quark
	//			%M - membuf (only for text)

	virtual Result prepare(char* queryText, Bool useAsIs = FALSE);
	//		- prepares dsql query with associated parameters, where type
	//		specifiers preceeding by '%' must be used instead of '?' symbols
	//		Later, in start real parameters must be given according with the
	//		query string type specifiers
	//			For instance :
	//		prepare( "SELECT KUCH FROM SCHED WHERE AK=%S AND REIS=%S AND MOD=%S");
	//		start(NULL, &fltKey.ak, &fltKey.flightNo, &fltKey.flightMod);

	virtual Bool unprepare();
	//		- drops preparing data to allow usage this object for
	//		another query. This function can be applied only for
	//		unnamed queries. Returns TRUE when done.

	virtual Result open(char* queryText, ...);
	//		- starts dsql select query, if query has not been prepared yet,
	//		it would be prepared using queryText, so if queryText = NULL, then
	// 	query must be prepared before call of open
	//		The list of parameters must contain actual values for executing
	//		the query and must be consistent with the types pointed in queryText

	virtual Result get(char* queryText, ...) = 0;
	//		- executes query that returns a single row
	//		if query has not been prepared yet,
	//		it would be prepared using queryText, so if queryText = NULL, then
	// 	query must be prepared before call of open
	//		The list of parameters must contain actual values for executing
	//		the query and must be consistent with the types pointed in queryText

	virtual Result executeImmediate(char *queryText) = 0;
	virtual Result execute(char* queryText, ...);
	//		- executes non-select query and works as open

	virtual Result read(char* format, ...);
	//		- reads current record buffer into variables, pointed in the
	//		argument list, according with the given format.
	//		format must have one of the following forms :
	//		1)	"%[<type1>]%[<type2>]...%[<typeN>]",
	//		2)	"%[<type1>](<field1>)...%[<typeN>](<fieldN>)",
	//		3)	"%[<type1>]#<fieldNo1>...%[<typeN>]#<fieldNoN>"
	//		Here :
	//			<typeI> - is the type specifier, which are the same as in queryText.
	//			<fieldI> - the name of the field in database
	//			<fieldNoI> - the ordinal no of the field in query
	//		In all forms <type> is optional and can be omitted, that causes
	//		to interprete the according variable having the default type for
	//		the given field.

	virtual Bool fetch();
	virtual Bool fetch(char* format, ...);
	//		- execute usual fetch, retrieving the next record of the query,
	//		and then call read with the same arguments

	//		When start is executing, the variables of character based types are
	//		converting into the DATABASE codepage, and during the read, all
	//    character based data from record buffer will be converted to
	//		the DEFAULT codepage before writing into variables.
	//		The similar operations take a place for Date type variables.

	virtual Result close(Bool ignoreError = FALSE) = 0;
	virtual Result drop(Bool ignoreError = FALSE) = 0;
	static Result dropQuery(const char* name);  	//		Drops stored query
	static Result dropStack(const char* name);	//		Drops all stack


	// Field access methods
	virtual char* getFieldName(int fldNo);
	virtual int getFieldNo(const char *fldName);
	virtual int getFieldType(int fldNo);
	virtual int getFieldType(const char *fldName);
	virtual int getFieldLength(int fldNo);
	virtual int getFieldLength(const char *fldName);
	virtual Bool isFieldNull(int fldNo);
	virtual Bool isFieldNull(const char *fldName);
	virtual Pointer getFieldValue(int fldNo);
	virtual Pointer getFieldValue(const char *fldName);
	virtual char* getTextField(int fldNo);


//	virtual Bool isChar(int fldNo) = 0;
//	virtual Bool isBlob(int fldNo) = 0;

//	char* recbuf;

	char queryName[QNAME_LEN];
	short format[MAX_FIELDS_IN_QUERY];

	short state;
	DbSQLDA in;
	DbSQLDA out;
	SQLDA* out_sqlda;		//		It never points to allocated data

	DynArray* fieldByNameAccelerator;
	static short nullind;


	static short typeToShort(char* ptr, int& size);
	void getFormat(char* queryText);

//	void allocateSQLDAbuffer();

	Result store(char* text); 			//	Store the query
	Bool restore(char* text);			//	Try to find applicable stored query and apply it
	virtual void cloneFrom(const __Query& query);
	virtual __Query* dup() const = 0;

//	virtual void fillInSQLDA(MemBuf& parBuffer, va_list ap) = 0;

//	virtual preOpen(char* queryText, short method = 0);

	virtual Result _prepare(char* queryText) = 0;
	virtual Result _unprepare();
	virtual _open(SQLDA* _in_sqlda = 0, short method = 0) = 0;
	virtual _execute(SQLDA* _in_sqlda = 0, short method = 0) = 0;
	virtual Bool _fetch() = 0;
	virtual Result _read(char* format, va_list ap);
	virtual Result openCursor() = 0;

//	virtual Result processInArg(MemBuf& dest, short type, SQLVAR* var, va_list& ap);
//	Result _readOut(void* data, short type, int fieldNo);

	char* preRead(short& fieldNo, short& type, char* ptr);
	void fillSQLDA(DbSQLDA& sqlda, va_list& ap);

	virtual Result setArg(MemBuf& dest, short type, SQLVAR* var, va_list& ap);
	virtual Result getArg(DbSQLDA& sqlda, void* data, short type, int i);

	virtual void printField(DbSQLDA& s, int i, MemBuf& dest, char* format = 0);
	void print(DbSQLDA& s, MemBuf& dest);

//	virtual Result describeInput(DbSQLDA& sqlda) = 0;
//	virtual Result describeOutput(DbSQLDA& sqlda) = 0;
	virtual Result getUniArg(DbSQLDA& sqlda, int i, UniVar* uv);
	virtual Result setUniArg(DbSQLDA& sqlda, int i, UniVar* uv);

	virtual Result play(Dump& data, char* queryText, ...);
	virtual Result replay(Dump& data);
	};

class StatStarter	{
public :
	Statistics* stat;

	StatStarter()	{	stat = 0;	}
	void init(char* name)	{
		stat = Statistics::find(name);
		if (stat)
			stat->startRequest();
		}
	void stop()	{	if (stat)	stat->stopRequest();	}
	~StatStarter()	{	stop();	}
	};

#endif
